# License: BSD 3 clause
"""
Easily run common scikit-learn experiments with pre-generated features.

:author: Nitin Madnani (nmadnani@ets.org)
:author: Dan Blanchard (dblanchard@ets.org)
:author: Michael Heilman (mheilman@ets.org)
:organization: ETS
"""
